import os
import unittest
from importlib import reload

from confyml import confyml

from tests.resources import sample_module

FILE_PATH = os.path.dirname(__file__)


class TestConfigMethodsMode3(unittest.TestCase):

    def setUp(self) -> None:
        file_path = os.path.dirname(__file__)
        yaml_file = os.path.abspath(os.path.join(
                                     file_path,
                                     'resources/1.test.config.yaml'))
        self._sut = confyml.Config(yaml_file, mode='mcf')

    def test_object_is_config(self):
        self.assertIsInstance(self._sut, confyml.Config)

    def test_should_return_list_of_configured_objects(self):
        objects = ['example_module1.ExampleClass11',
                   'example_module1.ExampleClass11.example_function111',
                   'example_module1.ExampleClass12',
                   'example_module1.ExampleClass12.example_function121',
                   'example_module1.ExampleClass12.example_function122',
                   'example_module_2.ExampleClass21.example_function211']
        self.assertEqual(set(objects), set(self._sut._get_objects()))

    def test_should_return_config_for_object(self):
        actual = self._sut._get_config('example_module1.ExampleClass11')
        expt = {'kwarg_1101': 1, 'kwarg_1102': 'test'}
        self.assertDictEqual(expt, actual)
        actual = self._sut._get_config('example_module1.ExampleClass11'
                                       '.example_function111')
        expt = {'kwarg_1111': 2, 'kwarg_1112': 'function_test',
                'kwarg_1113': 'yaml_test'}
        self.assertDictEqual(expt, actual)


class TestConfigMethodsMode2(unittest.TestCase):

    def setUp(self) -> None:
        file_path = os.path.dirname(__file__)
        yaml_file = os.path.abspath(os.path.join(
                                     file_path,
                                     'resources/2.test.config.yaml'))
        self._sut = confyml.Config(yaml_file, mode='cf')

    def test_object_is_config(self):
        self.assertIsInstance(self._sut, confyml.Config)

    def test_should_return_list_of_configured_objects(self):
        objects = ['ExampleClass', 'ExampleClass.example_function']
        self.assertEqual(set(objects), set(self._sut._get_objects()))

    def test_should_return_config_for_object(self):
        conf = self._sut._get_config('ExampleClass.example_function')
        expt = {'kwarg_1': 2,
                'kwarg_2': 'function_test',
                'kwarg_3': 'yaml_test'}
        self.assertDictEqual(expt, conf)
        conf = self._sut._get_config('ExampleClass')
        expt = {'kwarg_1': 1, 'kwarg_2': 'test'}
        self.assertDictEqual(expt, conf)

    def test_should_apply_config(self):
        confyml.set_config('resources/2.sample.config.yaml', 'cf')
        reload(sample_module)
        ret = sample_module.SampleClass().sample_method()
        self.assertEqual((2, 1), ret)
        ret = sample_module.sample_function()
        self.assertEqual(1, ret)


class TestConfigMethodsMode1(unittest.TestCase):

    def setUp(self) -> None:
        objects = os.path.abspath(os.path.join(
                                     FILE_PATH,
                                     'resources/3.test.config.yaml'))
        self._sut = confyml.Config(objects, mode='f')

    def test_object_is_config(self):
        self.assertIsInstance(self._sut, confyml.Config)

    def test_should_return_list_of_configured_objects(self):
        objects = ['example_function']
        self.assertEqual(set(objects), set(self._sut._get_objects()))

    def test_should_return_config_for_object(self):
        actual = self._sut._get_config('example_function')
        expt = {'kwarg_1': 2,
                'kwarg_2': 'function_test',
                'kwarg_3': 'yaml_test'}
        self.assertDictEqual(expt, actual)

    def test_should_apply_config(self):
        confyml.set_config('resources/sample.config.yaml', 'mcf')
        reload(sample_module)
        ret = sample_module.SampleClass().sample_method()
        self.assertEqual((2, 1), ret)
        ret = sample_module.sample_function()
        self.assertEqual(1, ret)

    def test_should_not_apply_config(self):
        os.environ['CONFYML_CONFIG'] = ''
        reload(sample_module)
        ret = sample_module.SampleClass().sample_method()
        self.assertEqual((1, None), ret)
        ret = sample_module.sample_function()
        self.assertEqual(None, ret)

    def test_should_be_overwritten_by_call(self):
        confyml.set_config('resources/sample.config.yaml', 'mcf')
        reload(sample_module)
        ret = sample_module.sample_function_2(kwarg1=2)
        self.assertEqual(2, ret)
        ret = sample_module.sample_function_3()
        self.assertIsNone(ret)



from confyml import confyml

config = confyml.get_config()


class SampleClass:

    @config.apply
    def __init__(self, kwarg1=None):
        self._kwarg1 = kwarg1

    @config.apply
    def sample_method(self, kwarg1=1):
        return kwarg1, self._kwarg1


@config.apply
def sample_function(kwarg1=None):
    return kwarg1


@config.apply
def sample_function_2(kwarg1=None):
    return kwarg1


@config.apply
def sample_function_3(kwarg1=None):
    return kwarg1


import os
import traceback
import warnings
from functools import wraps
from typing import List, Tuple, Set, Dict

import yaml


class Config:

    def __init__(self, yaml_file: str, mode: str = 'mcf') -> None:
        """
        Instantiate the Config class

        Parameters
        ----------
        yaml_file : str
            File path to yaml config file
        mode : str
            Mode that config is being applied
              'mcf' : modules and classes and methods and/or modules .
                      and functions
              'mf' : modules and functions only.
              'cf' : classes and methods and/or functions.
              'f' : functions only.
        """
        self._path = yaml_file
        self._mode_name = mode
        mode_map = {'mcf': 3, 'mf': 2, 'cf': 2, 'f': 1}
        self._mode = mode_map[mode]
        self._config = {}
        if yaml_file:
            with open(yaml_file, 'r') as f:
                self._config = yaml.safe_load(f)

        self._configured = self._get_objects()

    @property
    def mode(self) -> int:
        return self._mode

    def __getitem__(self, item: str):
        return self._config[item]

    def _get_objects(self) -> List[Set]:
        """
        Gets a list of all functions and classes that have been
        configured in the yaml config.

        Returns
        -------
        set
            Set of object names that are configured in the yaml.
        """
        named = (None, self._config.copy())

        out, log = _reduce([named])

        for i in range(self.mode):
            out, log = _reduce(out, log)
        return list(set(log))

    def _get_config(self, name) -> Dict:
        conf = self._config.copy()
        key_list = name.split('.')
        for k in key_list:
            conf = conf[k]
        return {k: v for k, v in conf.items() if not isinstance(v, dict)}

    def apply(self, func: object) -> object:
        """Apply the config to the function."""
        @wraps(func)
        def wrapper(*args, **kwargs):
            name_map = {'f': func.__name__,
                        'mf': f'{func.__module__.split(".")[-1]}'
                              f'.{func.__name__}',
                        'cf': func.__qualname__,
                        'mcf': f'{func.__module__.split(".")[-1]}'
                               f'.{func.__qualname__}'}

            name = name_map[self._mode_name]
            if '.__init__' in name:
                name = name.replace('.__init__', '')
            return self._wrap(name, func, *args, **kwargs)
        return wrapper

    def _wrap(self, name, func, *args, **kwargs):
        if name not in self._configured:
            return func(*args, **kwargs)

        func_to_replace = [fn for fn in self._configured if fn == name]

        if len(func_to_replace) > 1:
            raise ReferenceError(f'Multiple definitions for function {name}')

        conf = self._get_config(func_to_replace[0])
        conf.update(**kwargs)
        return func(*args, **conf)


def _reduce(ds: List[Tuple], store: List = None) -> Tuple[List[Tuple], List]:
    """
    Returns list of tuples and list of key names associated. Not
    to be used directly.
    """
    # Store to hold the parent key names
    if store is None:
        store = []

    # Loop through _NamedDicts in the list
    dicts = []
    for d in ds:
        for k, v in d[1].items():
            if isinstance(v, dict):

                # If item is a dict itself, create a _NamedDict and
                # add to the name with the key of that dict.
                new_name = d[0] + '.' + k if d[0] else k
                inner_d = (new_name, v)

                dicts.append(inner_d)

            elif not isinstance(v, dict):
                # If the value is a string, not a dict, this is the
                # level the kwargs are set
                store.append(d[0])

    return dicts, store


def set_config(yaml_path, mode='mcf'):
    stack = traceback.extract_stack()
    caller = os.path.dirname(stack[-2].filename)
    full_path = os.path.realpath(os.path.join(caller, yaml_path))

    os.environ['CONFYML_CONFIG'] = str(full_path)
    os.environ['CONFYML_MODE'] = str(mode)

    try:
        get_config()
    except FileNotFoundError as e:
        del os.environ['CONFYML_CONFIG']
        del os.environ['CONFYML_MODE']
        raise FileNotFoundError("Config YML not found")


def get_config():
    file = os.environ.get('CONFYML_CONFIG')
    mode = os.environ.get('CONFYML_MODE')

    if file is None:
        warnings.warn("Confyml Config imported without config file set. Set "
                      "environment variable CONFYML_CONFIG to apply "
                      "config.", UserWarning, stacklevel=2)
    if mode:
        return Config(file, mode=mode)
    return Config(file)



import codecs
import os.path

from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

version = get_version('confyml/__init__.py')

setup(
    name="confyml",
    version=version,
    author="iccyp",
    author_email='iccyp.py@gmail.com',
    description="Small package to apply YAML config to methods and functions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/iccyp/confyml",
    download_url='https://github.com/iccyp/confyml/archive/'
                 f'{version}.tar.gz',
    packages=find_packages(),
    install_requires=['PyYAML'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
